##########################################################################
# configurator - simple config program for midori linux
# written by Claudio Fleiner
#
# Copyright (C) 2000 Transmeta Corporation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Note: paramter by itself can be used to create paramters, but as soon
# as the type is known it gets reblessed as a subclass

# manages individual parameters. This package is not meant to be
# used directly, instead one of its children should be used.
package PARAMETER;

# List of known parameter types.
my @paramtypes;

# register a new parameter type
sub register {
   my $param=shift;
   push @paramtypes,$param;
}

# Create a new parameter
sub new {
   my $p=shift;
   my $self;
   $self={ };
   bless $self;
   $self->name($p);
   $self->ml(shift);
   return $self;
}

# name of the parameter
sub name {
   my $self = shift;
   if (@_) { $self->{NAME} = shift }
   return $self->{NAME};
}

# package to which this parameter belongs too
sub ml {
   my $self = shift;
   if (@_) { $self->{ML} = shift }
   return $self->{ML};
}

# value of the parameter in the config file
sub config {
   my $self = shift;
   if (@_) { $self->{CONFIG} = shift }
   return $self->{CONFIG};
}

# current value of this parameter
sub edit {
   my $self = shift;
   if (@_) { 
      $self->{EDIT} = shift;
      $self->{CURRENT} = $self->{EDIT};
   }
   return $self->{EDIT};
}

# default value (as defined in the CONFIG file)
sub default {
   my $self = shift;
   if (@_) { $self->{DEFAULT} = shift }
   return $self->{DEFAULT};
}

# true if value in config file is different than
# default value
sub changed {
   my $self = shift;
   return $self->config ne $self->default;
}

# true if edited value differnet that config value
sub edited {
   my $self = shift;
   return $self->edit ne $self->config;
}

# comment (from config file)
sub comment {
   my $self = shift;
   if (@_) { $self->{COMMENT} = shift }
   return $self->{COMMENT};
}


# type used when creating the parameter
sub realtype {
   my $self = shift;
   if (@_) { $self->{REALTYPE} = shift }
   return $self->{REALTYPE};
}

# known types
sub type {
   my $self = shift;
   if (@_) { 
      my $type=shift;
      if(defined($type) && $type ne "")  {
         T: {
            foreach my $i (@paramtypes) {
               last T if eval("${i}::settype(\$self,\$type)");
            }
            print "unknown type '$type'\n";
         }
      }
   }
   if ((not defined $self->{TYPE}) or ($self->{TYPE} eq "")) {
      if($self->default eq "yes" || $self->default eq "no") { $self->{TYPE}=("YESNO"); }
      elsif($self->default eq "true" || $self->default eq "false") { $self->{TYPE}=("TRUEFALSE"); }
      elsif($self->default eq "YES" || $self->default eq "NO") { $self->{TYPE}=("YESNO"); }
      elsif($self->default eq "TRUE" || $self->default eq "FALSE") { $self->{TYPE}=("TRUEFALSE"); }
      else { PARAM_REGEX->settype($self,"STRING"); }
   }
   return $self->{TYPE};
}

# Check if the value to which this parameter is set is acceptable
sub check { return 1; }

# with no paramter check if $self->edit is an acceptable value
# otherwise check parameter
sub check_value {
   my $self=shift;
   my $ch=shift;
   $ch=$self->edit if(!defined($ch));
   return $self->check($ch);
}

# Check if the type passed in is the correct one for a given 
# subclass. If so bless self.
# $1: seld
# $2: type string (from CONFIG file)
sub settype {
   die "settype needs to be redifined in every subclass\n";
}

##########################################################################
## Readonly parameters (cannot be changed via the webconfig interface)
package PARAM_READONLY;

PARAMETER::register("PARAM_READONLY");
@ISA=qw( PARAMETER );

sub settype {
   my $self=shift;
   my $realtype=shift;
   my $type=$realtype;


   $type="READONLY" if($realtype =~ "READONLY.*");

   return 0 if($type ne "READONLY");
   bless $self;

   $self->{TYPE}="READONLY";
   $self->realtype($realtype);
}



##########################################################################
# Hidden parameters are not visible in the webconfig interface
package PARAM_HIDDEN;

PARAMETER::register("PARAM_HIDDEN");
@ISA=qw( PARAMETER );

sub settype {
   my $self=shift;
   my $realtype=shift;
   my $type=$realtype;


   $type="HIDDEN" if($realtype =~ "HIDDEN.*");

   return 0 if($type ne "HIDDEN");
   bless $self;

   $self->{TYPE}="HIDDEN";
   $self->realtype($realtype);
}


##########################################################################
# REGEX is used for all types that expect a simple string as input. Even
# STRING, INTEGER and FLOAT use this package
package PARAM_REGEX;

PARAMETER::register("PARAM_REGEX");
@ISA=qw( PARAMETER );

sub settype {
   my $self=shift;
   my $realtype=shift;
   my $type=$realtype;


   ($type,$regex)=("REGEX",".*") if($realtype eq "STRING");
   ($type,$regex)=("REGEX",'^\d+$') if($realtype eq "INTEGER");
   ($type,$regex)=("REGEX",'^(\.\d+)|(\d+(.\d*)?)$') if($realtype eq "FLOAT");
   ($type,$regex)=("REGEX",$1) if($realtype =~ /^REGEX (.*)/s);
   ($type,$regex)=("REGEX",".*") if($realtype eq "REGEX");

   return 0 if($type ne "REGEX");
   bless $self;

   if(!defined($regex)) {
      $regex=".*";
      print "PARAM: ".$self->name." ???? type REGEX with no regular expression, assuming '.*'\n";
   }

   eval (my $x="test" =~ /$regex/);

   if($@) {
      print "PARAM: ".$self->name." regular expression error for '$regex': $@\n";
      print "                        using '.*' instead.\n";
      $regex=".*";
   }
   $self->regex($regex);
   $self->{TYPE}="REGEX";
   $realtype=~s/(^\S+).*/$1/;
   $self->realtype($realtype);
}

# optionally set regular expression to be used.
# Return currently used regex
sub regex {
   my $self = shift;
   if (@_) { $self->{REGEX} = shift }
   return $self->{REGEX};
}

# check if $self->edit is an acceptable value
sub check {
   my $self=shift;
   my $ch=shift;
   my $r=$self->regex;
   return $ch =~ /$r/;
}

##########################################################################
# ENUM lets the user select ONE entry from a list
package PARAM_ENUM;

@ISA=qw( PARAMETER );
PARAMETER::register("PARAM_ENUM");

sub settype {
   my $self=shift;
   my $realtype=shift;
   my $type=$realtype;
   my $enum="";
   my $radio=0;


   ($type,$enum)=("ENUM","yes,no") if($realtype eq "YESNO");
   ($type,$enum)=("ENUM","true,false") if($realtype eq "TRUEFALSE");
   ($type,$enum)=("ENUM","on,off") if($realtype eq "ONOFF");
   ($type,$enum)=("ENUM",$1) if($realtype =~ /^BOOLEAN\s*\((.*)\)\s*$/s);
   ($type,$enum)=("ENUM",$1) if($realtype =~ /^ENUM\s*\(\s*(.*)\)\s*$/s);
   ($type,$enum,$radio)=("ENUM",$1,1) if($realtype =~ /^ENUM\s+RADIO\s*\(\s*(.*)\)\s*$/s);


   return 0 if($type ne "ENUM");

   bless $self;

   if(!defined($enum)) {
      print "PARAM: ".$self->name." ???? type ENUM with no options?\n";
      return;
   }

   my $options={ };
   my $order=[];


T: while(length($enum)>0) {
      $enum=~s/^\s*//s;
      my $o="";
      my $d="";

      # get the option first
      if($enum =~ /^"([^"]*)"\s*(.*)/s) {
         $o=$1;
         $enum=$2;
      } elsif($enum =~ /^'([^"]*)'\s*(.*)/s) {
         $o=$1;
         $enum=$2;
      } elsif($enum =~/^([^:,]*[^ \t:,])\s*(.*)/s) {
         $o=$1;
         $enum=$2;
      } else {
         print "ENUM ERROR OPTION: '$enum'\n";
         last T;
      } 

      # get the description if existant
      if($enum =~ /^:\s*(.*)/s) {
         $enum=$1;
         if($enum =~ /^"([^"]*)"\s*(.*)/s) {
            $d=$1;
            $enum=$2;
         } elsif($enum =~ /^'([^']*)'\s*(.*)/s) {
            $d=$1;
            $enum=$2;
         } elsif($enum =~/^([^,]*[^, \t])\s*(.*)/s) {
            $d=$1;
            $enum=$2;
         } else {
            print "ENUM ERROR DESCRIPTION: '$enum'\n";
            last T;
         } 
      }

      $$options{$o}=$d;
      push @$order,$o;
      if($enum =~ /^,\s*(.*)/s) {
         $enum=$1;
      } else {
         if($enum !~ /^\s*$/s) {
            print "ENUM ERROR, expected ',': '$enum'\n";
         }
         last T;
      }
   }

   $self->radio($radio);
   $self->options($options);
   $self->optionsorder($order);
   $self->{TYPE}="ENUM";
   $realtype=~s/(^\S+).*/$1/;
   $self->realtype($realtype);
}

# return list of options (hash with NAME->string to print)
sub options {
   my $self = shift;
   if (@_) { $self->{OPTIONS} = shift }
   return $self->{OPTIONS};
}

# return true if radio buttons should be used instead of a list
sub radio {
   my $self = shift;
   if (@_) { $self->{RADIO} = shift }
   return $self->{RADIO};
}

# return the options in the same order as they are in the CONFIG file
sub optionsorder {
   my $self = shift;
   if (@_) { $self->{OPTIONSORDER} = shift }
   return $self->{OPTIONSORDER};
}


# check if $self->edit is an acceptable value
sub check {
   my $self=shift;
   my $o=$self->options;
   my $ch=shift;
   return defined($$o{$ch});
}

##########################################################################
# LIST:  let the user select 0 or more values from a list
package PARAM_LIST;

@ISA=qw( PARAM_ENUM );
PARAMETER::register("PARAM_LIST");

sub settype {
   my $self=shift;
   my $realtype=shift;
   my $type=$realtype;
   my $enum="";
   my $radio=0;


   ($type,$enum)=("LIST",$1) if($realtype =~ /^LIST\s*\(\s*(.*)\)\s*$/s);
   ($type,$enum,$radio)=("LIST",$1,1) if($realtype =~ /^LIST\s+CHECKBOX\s*\(\s*(.*)\)\s*$/s);

   return 0 if($type ne "LIST");

   bless $self;

   if(!defined($enum)) {
      print "PARAM: ".$self->name." ???? type LIST with no options?\n";
      return;
   }

   my $options={ };
   my $order=[];


T: while(length($enum)>0) {
     $enum=~s/^\s*//s;
      my $o="";
      my $d="";

      # get the option first
      if($enum =~ /^"([^" \t]*)"\s*(.*)/s) {
         $o=$1;
         $enum=$2;
      } elsif($enum =~ /^'([^" \t]*)'\s*(.*)/s) {
         $o=$1;
         $enum=$2;
      } elsif($enum =~/^([^:,]*[^ \t,:])\s*(.*)/s) {
         $o=$1;
         $enum=$2;
      } else {
         print "LIST ERROR OPTION: '$enum'\n";
         last T;
      } 

      # get the description if existant
      if($enum =~ /^:\s*(.*)/s) {
         $enum=$1;
         if($enum =~ /^"([^"]*)"\s*(.*)/s) {
            $d=$1;
            $enum=$2;
         } elsif($enum =~ /^'([^']*)'\s*(.*)/s) {
            $d=$1;
            $enum=$2;
         } elsif($enum =~/^([^,]*[^ \t,:])\s*(.*)/s) {
            $d=$1;
            $enum=$2;
         } else {
            print "LIST ERROR DESCRIPTION: '$enum'\n";
            last T;
         } 
      }
      $$options{$o}=$d;
      push @$order,$o;
      if($enum =~ /^,\s*(.*)/s) {
         $enum=$1;
      } else {
         if($enum !~ /^\s*$/s) {
            print "LIST ERROR, expected ',': '$enum'\n";
         }
         last T;
      }
   }

   $self->radio($radio);
   $self->options($options);
   $self->optionsorder($order);
   $self->{TYPE}="LIST";
   $realtype=~s/(^\S+).*/$1/;
   $self->realtype($realtype);
}

# check if $self->edit is an acceptable value
sub check {
   my $self=shift;
   my $o=$self->options;
   my $ch=shift;
   foreach my $i (split(/ +/,$ch)) {
      return 0 if(!defined($$o{$i}));
   }
   return 1;
}

##########################################################################
# PERL package, no special treatment as everything is done the main package
# already.
package PARAM_PERL;

PARAMETER::register("PARAM_PERL");

sub settype {
   my $self=shift;
   my $realtype=shift;
   my $type=$realtype;
   my $perl="";


   ($type,$perl)=("PERL",$1) if($realtype =~ /^PERL\s*(.*)$/s);

   return 0 if($type ne "PERL");

   if(!eval($perl)) {
      print "error evaluting perl code: ".$@."\n";
      print "package: ".$self->ml->name.", parameter: ".$self->name."\n";
      exit 1;
      return 0;
   }

   if(!eval("bless \$self, ".$self->name."\n")) {
      print "error blessing code\n";
      return 0;
   }

   $self->initialize;
   $self->{TYPE}="PERL";
   return 1;
}

1;
